@@ -352,7 +352,9 @@ bin_zle(char *name, char **args, Options ops, UNUSED(int func))
 	{ 'U', bin_zle_unget, 1, 1 },
 	{ 'K', bin_zle_keymap, 1, 1 },
 	{ 'I', bin_zle_invalidate, 0, 0 },
+	{ 'f', bin_zle_flags, 1, -1 },
 	{ 'F', bin_zle_fd, 0, 2 },
+	{ 'T', bin_zle_transform, 0, 2},
 	{ 0,   bin_zle_call, 0, -1 },
     };
     struct opn const *op, *opp;
@@ -394,9 +396,13 @@ bin_zle_list(UNUSED(char *name), char **args, Options ops, UNUSED(char func))
 	Thingy t;
 
 	for (; *args && !ret; args++) {
-	    if (!(t = (Thingy) thingytab->getnode2(thingytab, *args)) ||
+	    HashNode hn = thingytab->getnode2(thingytab, *args);
+	    if (!(t = (Thingy) hn) ||
 		(!OPT_ISSET(ops,'a') && (t->widget->flags & WIDGET_INT)))
 		ret = 1;
+	    else if (OPT_ISSET(ops,'L')) {
+		scanlistwidgets(hn, 1);
+	    }
 	}
 	return ret;
     }
@@ -461,7 +467,7 @@ bin_zle_mesg(char *name, char **args, UNUSED(Options ops), UNUSED(char func))
 static int
 bin_zle_unget(char *name, char **args, UNUSED(Options ops), UNUSED(char func))
 {
-    char *b = *args, *p = b + strlen(b);
+    char *b = unmeta(*args), *p = b + strlen(b);
 
     if (!zleactive) {
 	zwarnnam(name, "can only be called from widget function");
@@ -483,6 +489,12 @@ bin_zle_keymap(char *name, char **args, UNUSED(Options ops), UNUSED(char func))
     return selectkeymap(*args, 0);
 }
 
+/*
+ * List a widget.
+ * If list is negative, just print the name.
+ * If list is 0, use abbreviated format.
+ * If list is positive, output as a command.
+ */
 /**/
 static void
 scanlistwidgets(HashNode hn, int list)
@@ -614,7 +626,7 @@ bin_zle_complete(char *name, char **args, UNUSED(Options ops), UNUSED(char func)
 
 /**/
 static int
-zle_usable()
+zle_usable(void)
 {
     return zleactive && !incompctlfunc && !incompfunc
 #if 0
@@ -629,6 +641,48 @@ zle_usable()
 	   ;
 }
 
+/**/
+static int
+bin_zle_flags(char *name, char **args, UNUSED(Options ops), UNUSED(char func))
+{
+    int ret = 0;
+    char **flag;
+
+    if (!zle_usable()) {
+	zwarnnam(name, "can only set flags from a widget");
+	return 1;
+    }
+
+    if (bindk) {
+	Widget w = bindk->widget;
+	if (w) {
+	    for (flag = args; *flag; flag++) {
+		if (!strcmp(*flag, "yank")) {
+		    w->flags |= ZLE_YANKAFTER;
+		} else if (!strcmp(*flag, "yankbefore"))
+		    w->flags |= ZLE_YANKBEFORE;
+		else if (!strcmp(*flag, "kill"))
+		    w->flags |= ZLE_KILL;
+		/*
+		 * These won't do anything yet, because of how execzlefunc
+		 * handles user widgets
+		} else if (!strcmp(*flag, "menucmp"))
+		    w->flags |= ZLE_MENUCMP;
+		else if (!strcmp(*flag, "linemove"))
+		    w->flags |= ZLE_LINEMOVE;
+		else if (!strcmp(*flag, "keepsuffix"))
+		    w->flags |= ZLE_KEEPSUFFIX;
+		*/
+		else {
+		    zwarnnam(name, "invalid flag `%s' given to zle -f", *flag);
+		    ret = 1;
+		}
+	    }
+	}
+    }
+    return ret;
+}
+
 /**/
 static int
 bin_zle_call(char *name, char **args, UNUSED(Options ops), UNUSED(char func))
@@ -767,10 +821,12 @@ bin_zle_fd(char *name, char **args, Options ops, UNUSED(char func))
 	    return 1;
 	}
 	for (i = 0; i < nwatch; i++) {
-	    if (*args && watch_fds[i] != fd)
+	    Watch_fd watch_fd = watch_fds + i;
+	    if (*args && watch_fd->fd != fd)
 		continue;
 	    found = 1;
-	    printf("%s -F %d %s\n", name, watch_fds[i], watch_funcs[i]);
+	    printf("%s -F %s%d %s\n", name, watch_fd->widget ? "-w " : "",
+		   watch_fd->fd, watch_fd->func);
 	}
 	/* only return status 1 if fd given and not found */
 	return *args && !found;
@@ -781,9 +837,11 @@ bin_zle_fd(char *name, char **args, Options ops, UNUSED(char func))
 	char *funcnam = ztrdup(args[1]);
 	if (nwatch) {
 	    for (i = 0; i < nwatch; i++) {
-		if (watch_fds[i] == fd) {
-		    zsfree(watch_funcs[i]);
-		    watch_funcs[i] = funcnam;
+		Watch_fd watch_fd = watch_fds + i;
+		if (watch_fd->fd == fd) {
+		    zsfree(watch_fd->func);
+		    watch_fd->func = funcnam;
+		    watch_fd->widget = OPT_ISSET(ops,'w') ? 1 : 0;
 		    found = 1;
 		    break;
 		}
@@ -792,46 +850,38 @@ bin_zle_fd(char *name, char **args, Options ops, UNUSED(char func))
 	if (!found) {
 	    /* zrealloc handles NULL pointers, so OK for first time through */
 	    int newnwatch = nwatch+1;
-	    watch_fds = (int *)zrealloc(watch_fds, 
-					newnwatch * sizeof(int));
-	    watch_funcs = (char **)zrealloc(watch_funcs,
-					    (newnwatch+1) * sizeof(char *));
-	    watch_fds[nwatch] = fd;
-	    watch_funcs[nwatch] = funcnam;
-	    watch_funcs[newnwatch] = NULL;
+	    Watch_fd new_fd;
+	    watch_fds = (Watch_fd)zrealloc(watch_fds, 
+					   newnwatch * sizeof(struct watch_fd));
+	    new_fd = watch_fds + nwatch;
+	    new_fd->fd = fd;
+	    new_fd->func = funcnam;
+	    new_fd->widget = OPT_ISSET(ops,'w') ? 1 : 0;
 	    nwatch = newnwatch;
 	}
     } else {
 	/* Deleting a handler */
 	for (i = 0; i < nwatch; i++) {
-	    if (watch_fds[i] == fd) {
+	    Watch_fd watch_fd = watch_fds + i;
+	    if (watch_fd->fd == fd) {
 		int newnwatch = nwatch-1;
-		int *new_fds;
-		char **new_funcs;
+		Watch_fd new_fds;
 
-		zsfree(watch_funcs[i]);
+		zsfree(watch_fd->func);
 		if (newnwatch) {
-		    new_fds = zalloc(newnwatch*sizeof(int));
-		    new_funcs = zalloc((newnwatch+1)*sizeof(char*));
+		    new_fds = zalloc(newnwatch*sizeof(struct watch_fd));
 		    if (i) {
-			memcpy(new_fds, watch_fds, i*sizeof(int));
-			memcpy(new_funcs, watch_funcs, i*sizeof(char *));
+			memcpy(new_fds, watch_fds, i*sizeof(struct watch_fd));
 		    }
 		    if (i < newnwatch) {
 			memcpy(new_fds+i, watch_fds+i+1,
-			       (newnwatch-i)*sizeof(int));
-			memcpy(new_funcs+i, watch_funcs+i+1,
-			       (newnwatch-i)*sizeof(char *));
+			       (newnwatch-i)*sizeof(struct watch_fd));
 		    }
-		    new_funcs[newnwatch] = NULL;
 		} else {
 		    new_fds = NULL;
-		    new_funcs = NULL;
 		}
-		zfree(watch_fds, nwatch*sizeof(int));
-		zfree(watch_funcs, (nwatch+1)*sizeof(char *));
+		zfree(watch_fds, nwatch*sizeof(struct watch_fd));
 		watch_fds = new_fds;
-		watch_funcs = new_funcs;
 		nwatch = newnwatch;
 		found = 1;
 		break;
@@ -846,6 +896,69 @@ bin_zle_fd(char *name, char **args, Options ops, UNUSED(char func))
     return 0;
 }
 
+/**/
+static int
+bin_zle_transform(char *name, char **args, Options ops, UNUSED(char func))
+{
+    /*
+     * -1: too few arguments
+     * 0: just right
+     * 1: too many arguments
+     * 2: first argument not recognised
+     */
+    int badargs = 0;
+
+    if (OPT_ISSET(ops,'L')) {
+	if (args[0]) {
+	    if (args[1]) {
+		badargs = 1;
+	    } else if (strcmp(args[0], "tc")) {
+		badargs = 2;
+	    }
+	}
+	if (!badargs && tcout_func_name) {
+	    fputs("zle -T tc ", stdout);
+	    quotedzputs(tcout_func_name, stdout);
+	    putchar('\n');
+	}
+    } else if (OPT_ISSET(ops,'r')) {
+	if (!args[0]) {
+	    badargs = -1;
+	} else if (args[1]) {
+	    badargs = 1;
+	} else if (tcout_func_name) {
+	    zsfree(tcout_func_name);
+	    tcout_func_name = NULL;
+	}
+    } else {
+	if (!args[0] || !args[1]) {
+	    badargs = -1;
+	    /* we've already checked args <= 2 */
+	} else {
+	    if (!strcmp(args[0], "tc")) {
+		if (tcout_func_name) {
+		    zsfree(tcout_func_name);
+		}
+		tcout_func_name = ztrdup(args[1]);
+	    } else {
+		badargs = 2;
+	    }
+	}
+    }
+
+    if (badargs) {
+	if (badargs == 2) {
+	    zwarnnam(name, "-T: no such transformation '%s'", args[0]);
+	} else {
+	    char *way = (badargs > 0) ? "many" : "few";
+	    zwarnnam(name, "too %s arguments for option -T", way);
+	}
+	return 1;
+    }
+
+    return 0;
+}
+
 /*******************/
 /* initialiasation */
 /*******************/
