@@ -535,7 +535,7 @@ lexconstant(void)
 	for (ptr2 = ptr; ptr2 < nptr; ptr2++) {
 	    if (*ptr2 == '_') {
 		int len = nptr - ptr;
-		ptr = strdup(ptr);
+		ptr = ztrdup(ptr);
 		for (ptr2 = ptr; len; len--) {
 		    if (*ptr2 == '_')
 			chuck(ptr2);
@@ -893,7 +893,6 @@ getcvar(char *s)
     return mn;
 }
 
-
 /**/
 static mnumber
 setmathvar(struct mathvalue *mvp, mnumber v)
